
//========================================================================================
//
//		Fun with prime numbers - the sieve of Eratosthenes v2.1 
//		by Dan Brown
//		www.tacticaltinkering.co.uk
//
//========================================================================================

#include <iostream>
#include <Windows.h>

#define MAX			15485863		//15485863  //179424673 //982451653 //2038074743
#define GetPos(n)	((n-3)/2)		// We're now only holding odd numbers in our array, so we need this macro to map each number to the char
									// which represents it - ignoring number 1, so the first element in the array corresponds to number 3

#if MAX > 4294967295				// If we can, use 32 bit integers (twice as fast on a 32 bit system)
typedef unsigned __int64 BigInt;
#else
typedef unsigned int BigInt;
#endif

using namespace std;

//-----------------------------------------------------------------------------------------
// Name: CTimer [CLASS]
// Desc: A class for use as a fairly accurate stopwatch
//-----------------------------------------------------------------------------------------
class CTimer	{
	private:
		__int64 start;
		__int64 end;
		__int64 iElapsed()
		{
			LARGE_INTEGER lInt;
			QueryPerformanceCounter(&lInt);
			return (lInt.QuadPart-start);
		}
	public:
		double value;
		double valueMS;
		bool Start()
		{
			LARGE_INTEGER lInt;
			if(!QueryPerformanceCounter(&lInt))
			{
				std::cout << "\nTimer failed to start!";
				return false;
			}
			start = lInt.QuadPart;
			return true;
		}
		bool Stop()
		{
			LARGE_INTEGER lIntPF, lInt;
			if(!QueryPerformanceFrequency(&lIntPF)) return false;
			if(!QueryPerformanceCounter(&lInt)) return false;			
			end = lInt.QuadPart-start;
			value = (double)iElapsed()/(double)lIntPF.QuadPart;
			valueMS = value*1000;
			return true;
		}
		double TimeElapsed()
		{
			Stop();
			return value;
		}
} Timer;
//-----------------------------------------------------------------------------------------

const BigInt arrSize = (int)((MAX-1)/2);	// Number of elements needed for all odd numbers from 3 up to MAX
char *n = new char[arrSize];				// Create the main array on the heap

int main()
{
	Timer.Start();

	std::fill(n, n + arrSize, 1);			// Assume all odd numbers (bar 1) are prime to start, so set all array values to 1

	cout << "\nInitialised the array (" << Timer.TimeElapsed() << "s)\n\nGenerating primes with the sieve of Eratosthenes ...\n\n";

	BigInt p = 3;			// p = current prime we're deleting multiples of in the bit array - ignoring 2!
	BigInt m = p*p;			// m = multiples of p greater than (p-1)*p, smaller multiples are already removed by smaller primes

	while(m <= MAX)
	{
		while(m <= MAX)							// Loop through the array and remove all multiples of p
		{
			n[GetPos(m)] = 0;
			m += 2*p;							// Ignore even multiples
		}
		
		do {p+=2;} while( n[GetPos(p)] != 1 );	// Let p be the next surviving number in the list
		m=p*p;									// Starting multiple of p
	}

	Timer.Stop();
	
	BigInt primesFound = 1;						// Not forgetting 2!
	for( BigInt a = 3 ; a <= MAX ; a+=2 )
		if(n[GetPos(a)]==1) 
		{
			primesFound++; 

			if(primesFound==1000000) cout << "\nMillionth prime:\t" << a;
			if(primesFound==10000000) cout << "\n10 millionth prime:\t" << a;
			if(primesFound==50000000) cout << "\n50 millionth prime:\t" << a;
			if(primesFound==100000000) cout << "\n100 millionth prime:\t" << a;
		}

	cout << "\n\nPrimes found:\t\t" << primesFound << "\nSearched up to:\t\t" << MAX << "\n\nTime elapsed:\t\t";
	Timer.value > 1 ? cout << Timer.value << "s" : cout << Timer.valueMS << "ms";
	
	cin.ignore();

	delete [] n;
}
